@@ -1,6 +1,6 @@
 /* vi:set et ai sw=2 sts=2 ts=2: */
 /*-
- * Copyright (c) 2007-2009 Jannis Pohlmann <jannis@xfce.org>
+ * Copyright (c) 2007-2010 Jannis Pohlmann <jannis@xfce.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -9,11 +9,11 @@
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU Library General Public
- * License along with this library; if not, write to the
+ * You should have received a copy of the GNU Library General
+ * Public License along with this library; if not, write to the
  * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  * Boston, MA 02110-1301, USA.
  */
@@ -28,7 +28,9 @@
 
 #include <glib/gi18n.h>
 
+#include <garcon/garcon-config.h>
 #include <garcon/garcon-environment.h>
+#include <garcon/garcon-marshal.h>
 #include <garcon/garcon-menu-element.h>
 #include <garcon/garcon-menu-item.h>
 #include <garcon/garcon-menu-directory.h>
@@ -38,7 +40,6 @@
 #include <garcon/garcon-menu-parser.h>
 #include <garcon/garcon-menu-merger.h>
 #include <garcon/garcon-private.h>
-#include <garcon/garcon.h>
 
 
 
@@ -97,7 +98,18 @@ enum
 
 
 
+/* Signal identifiers */
+enum
+{
+  RELOAD_REQUIRED,
+  DIRECTORY_CHANGED,
+  LAST_SIGNAL
+};
+
+
+
 static void                 garcon_menu_element_init                    (GarconMenuElementIface  *iface);
+static void                 garcon_menu_clear                           (GarconMenu              *menu);
 static void                 garcon_menu_finalize                        (GObject                 *object);
 static void                 garcon_menu_get_property                    (GObject                 *object,
                                                                          guint                    prop_id,
@@ -110,7 +122,9 @@ static void                 garcon_menu_set_property                    (GObject
 static void                 garcon_menu_set_directory                   (GarconMenu              *menu,
                                                                          GarconMenuDirectory     *directory);
 static void                 garcon_menu_resolve_menus                   (GarconMenu              *menu);
-static void                 garcon_menu_resolve_directory               (GarconMenu              *menu);
+static void                 garcon_menu_resolve_directory               (GarconMenu              *menu,
+                                                                         GCancellable            *cancellable,
+                                                                         gboolean                 recursive);
 static GarconMenuDirectory *garcon_menu_lookup_directory                (GarconMenu              *menu,
                                                                          const gchar             *filename);
 static void                 garcon_menu_collect_files                   (GarconMenu              *menu,
@@ -137,6 +151,43 @@ static const gchar         *garcon_menu_get_element_icon_name           (GarconM
 static gboolean             garcon_menu_get_element_visible             (GarconMenuElement       *element);
 static gboolean             garcon_menu_get_element_show_in_environment (GarconMenuElement       *element);
 static gboolean             garcon_menu_get_element_no_display          (GarconMenuElement       *element);
+static gboolean             garcon_menu_get_element_equal               (GarconMenuElement       *element,
+                                                                         GarconMenuElement       *other);
+static void                 garcon_menu_start_monitoring                (GarconMenu              *menu);
+static void                 garcon_menu_stop_monitoring                 (GarconMenu              *menu);
+static void                 garcon_menu_monitor_menu_files              (GarconMenu              *menu);
+static void                 garcon_menu_monitor_files                   (GarconMenu              *menu,
+                                                                         GList                   *files,
+                                                                         gpointer                 callback);
+static void                 garcon_menu_monitor_app_dirs                (GarconMenu              *menu);
+static void                 garcon_menu_monitor_directory_dirs          (GarconMenu              *menu);
+static void                 garcon_menu_file_changed                    (GarconMenu              *menu,
+                                                                         GFile                   *file,
+                                                                         GFile                   *other_file,
+                                                                         GFileMonitorEvent        event_type,
+                                                                         GFileMonitor            *monitor);
+static void                 garcon_menu_merge_file_changed              (GarconMenu              *menu,
+                                                                         GFile                   *file,
+                                                                         GFile                   *other_file,
+                                                                         GFileMonitorEvent        event_type,
+                                                                         GFileMonitor            *monitor);
+static void                 garcon_menu_merge_dir_changed               (GarconMenu              *menu,
+                                                                         GFile                   *file,
+                                                                         GFile                   *other_file,
+                                                                         GFileMonitorEvent        event_type,
+                                                                         GFileMonitor            *monitor);
+static void                 garcon_menu_app_dir_changed                 (GarconMenu              *menu,
+                                                                         GFile                   *file,
+                                                                         GFile                   *other_file,
+                                                                         GFileMonitorEvent        event_type,
+                                                                         GFileMonitor            *monitor);
+static void                 garcon_menu_directory_file_changed          (GarconMenu              *menu,
+                                                                         GFile                   *file,
+                                                                         GFile                   *other_file,
+                                                                         GFileMonitorEvent        event_type,
+                                                                         GFileMonitor            *monitor);
+static GarconMenuItem      *garcon_menu_find_file_item                  (GarconMenu              *menu,
+                                                                         GFile                   *file);
 
 
 
@@ -148,6 +199,13 @@ struct _GarconMenuPrivate
   /* DOM tree */
   GNode               *tree;
 
+  /* Merged menu files and merge directories */
+  GList               *merge_files;
+  GList               *merge_dirs;
+
+  /* File and directory monitors */
+  GList               *monitors;
+
   /* Directory */
   GarconMenuDirectory *directory;
 
@@ -162,12 +220,21 @@ struct _GarconMenuPrivate
 
   /* Shared menu item cache */
   GarconMenuItemCache *cache;
+
+  /* Flag for marking custom path menus */
+  guint                uses_custom_path : 1;
 };
 
 
 
 G_DEFINE_TYPE_WITH_CODE (GarconMenu, garcon_menu, G_TYPE_OBJECT,
-    G_IMPLEMENT_INTERFACE (GARCON_TYPE_MENU_ELEMENT, garcon_menu_element_init))
+                         G_IMPLEMENT_INTERFACE (GARCON_TYPE_MENU_ELEMENT, 
+                                                garcon_menu_element_init))
+
+
+
+static guint menu_signals[LAST_SIGNAL];
+static GQuark garcon_menu_file_quark;
 
 
 
@@ -211,6 +278,32 @@ garcon_menu_class_init (GarconMenuClass *klass)
                                                         GARCON_TYPE_MENU_DIRECTORY,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_STATIC_STRINGS));
+
+  menu_signals[RELOAD_REQUIRED] = 
+    g_signal_new ("reload-required",
+                  GARCON_TYPE_MENU,
+                  G_SIGNAL_RUN_LAST | G_SIGNAL_NO_HOOKS,
+                  0, 
+                  NULL,
+                  NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE,
+                  0);
+
+  menu_signals[DIRECTORY_CHANGED] =
+    g_signal_new ("directory-changed",
+                  GARCON_TYPE_MENU,
+                  G_SIGNAL_RUN_LAST | G_SIGNAL_NO_HOOKS,
+                  0,
+                  NULL,
+                  NULL,
+                  garcon_marshal_VOID__OBJECT_OBJECT,
+                  G_TYPE_NONE,
+                  2,
+                  GARCON_TYPE_MENU_DIRECTORY,
+                  GARCON_TYPE_MENU_DIRECTORY);
+
+  garcon_menu_file_quark = g_quark_from_string ("garcon-menu-file-quark");
 }
 
 
@@ -224,6 +317,7 @@ garcon_menu_element_init (GarconMenuElementIface *iface)
   iface->get_visible = garcon_menu_get_element_visible;
   iface->get_show_in_environment = garcon_menu_get_element_show_in_environment;
   iface->get_no_display = garcon_menu_get_element_no_display;
+  iface->equal = garcon_menu_get_element_equal;
 }
 
 
@@ -234,10 +328,14 @@ garcon_menu_init (GarconMenu *menu)
   menu->priv = GARCON_MENU_GET_PRIVATE (menu);
   menu->priv->file = NULL;
   menu->priv->tree = NULL;
+  menu->priv->merge_files = NULL;
+  menu->priv->merge_dirs = NULL;
+  menu->priv->monitors = NULL;
   menu->priv->directory = NULL;
   menu->priv->submenus = NULL;
   menu->priv->parent = NULL;
   menu->priv->pool = garcon_menu_item_pool_new ();
+  menu->priv->uses_custom_path = TRUE;
 
   /* Take reference on the menu item cache */
   menu->priv->cache = garcon_menu_item_cache_get_default ();
@@ -245,26 +343,61 @@ garcon_menu_init (GarconMenu *menu)
 
 
 
+static void
+garcon_menu_clear (GarconMenu *menu)
+{
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  /* Check if the menu is the root menu */
+  if (menu->priv->parent == NULL)
+    {
+      /* Stop monitoring recursively */
+      garcon_menu_stop_monitoring (menu);
+
+      /* Destroy the menu tree */
+      garcon_menu_node_tree_free (menu->priv->tree);
+      menu->priv->tree = NULL;
+
+      /* Release the merge files */
+      g_list_foreach (menu->priv->merge_files, (GFunc) g_object_unref, NULL);
+      g_list_free (menu->priv->merge_files);
+      menu->priv->merge_files = NULL;
+
+      /* Release the merge dirs */
+      g_list_foreach (menu->priv->merge_dirs, (GFunc) g_object_unref, NULL);
+      g_list_free (menu->priv->merge_dirs);
+      menu->priv->merge_dirs = NULL;
+    }
+
+  /* Free submenus */
+  g_list_foreach (menu->priv->submenus, (GFunc) g_object_unref, NULL);
+  g_list_free (menu->priv->submenus);
+  menu->priv->submenus = NULL;
+
+  /* Free directory */
+  if (G_LIKELY (menu->priv->directory != NULL))
+    {
+      g_object_unref (menu->priv->directory);
+      menu->priv->directory = NULL;
+    }
+
+  /* Clear the item pool */
+  garcon_menu_item_pool_clear (menu->priv->pool);
+}
+
+
+
 static void
 garcon_menu_finalize (GObject *object)
 {
   GarconMenu *menu = GARCON_MENU (object);
 
-  /* Destroy the menu tree */
-  if (menu->priv->parent == NULL)
-    garcon_menu_node_tree_free (menu->priv->tree);
+  /* Clear resources allocated in the load process */
+  garcon_menu_clear (menu);
 
   /* Free file */
   g_object_unref (menu->priv->file);
 
-  /* Free directory */
-  if (G_LIKELY (menu->priv->directory != NULL))
-    g_object_unref (menu->priv->directory);
-
-  /* Free submenus */
-  g_list_foreach (menu->priv->submenus, (GFunc) g_object_unref, NULL);
-  g_list_free (menu->priv->submenus);
-
   /* Free item pool */
   g_object_unref (menu->priv->pool);
 
@@ -410,26 +543,9 @@ GarconMenu *
 garcon_menu_new_applications (void)
 {
   GarconMenu *menu = NULL;
-  GFile      *file;
-  gchar      *filename;
-  guint       n;
 
-  /* Search for a usable applications menu file */
-  for (n = 0; menu == NULL && n < G_N_ELEMENTS (GARCON_MENU_ROOT_SPECS); ++n)
-    {
-      /* Search for the applications menu file */
-      filename = garcon_config_lookup (GARCON_MENU_ROOT_SPECS[n]);
-
-      /* Create menu if the file exists */
-      if (G_UNLIKELY (filename != NULL))
-        {
-          file = _garcon_file_new_for_unknown_input (filename, NULL);
-          menu = garcon_menu_new (file);
-          g_object_unref (file);
-        }
-
-      g_free (filename);
-    }
+  menu = g_object_new (GARCON_TYPE_MENU, NULL);
+  menu->priv->uses_custom_path = FALSE;
 
   return menu;
 }
@@ -547,20 +663,68 @@ garcon_menu_load (GarconMenu   *menu,
   GarconMenuMerger *merger;
   GHashTable       *desktop_id_table;
   gboolean          success = TRUE;
+  gchar            *filename;
+  guint             n;
 
   g_return_val_if_fail (GARCON_IS_MENU (menu), FALSE);
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
+  /* Make sure to reset the menu to a loadable state */
+  garcon_menu_clear (menu);
+
+  /* Check if we need to locate the applications menu file */
+  if (!menu->priv->uses_custom_path)
+    {
+      /* Release the old file if there is one */
+      if (menu->priv->file != NULL)
+        {
+          g_object_unref (menu->priv->file);
+          menu->priv->file = NULL;
+        }
+
+      /* Search for a usable applications menu file */
+      for (n = 0; 
+           menu->priv->file == NULL && n < G_N_ELEMENTS (GARCON_MENU_ROOT_SPECS);
+           ++n)
+        {
+          /* Search for the applications menu file */
+          filename = garcon_config_lookup (GARCON_MENU_ROOT_SPECS[n]);
+    
+          /* Use the file if it exists */
+          if (filename != NULL)
+            menu->priv->file = _garcon_file_new_for_unknown_input (filename, NULL);
+    
+          /* Free the filename string */
+          g_free (filename);
+        }
+
+      /* Abort with an error if no suitable applications menu file was found */
+      if (menu->priv->file == NULL)
+        {
+          g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_NOENT,
+                       _("No suitable application menu file found"));
+          return FALSE;
+        }
+    }
+
   parser = garcon_menu_parser_new (menu->priv->file);
 
-  if (G_LIKELY (garcon_menu_parser_run (parser, cancellable, error)))
+  if (garcon_menu_parser_run (parser, cancellable, error))
     {
       merger = garcon_menu_merger_new (GARCON_MENU_TREE_PROVIDER (parser));
 
-      if (garcon_menu_merger_run (merger, cancellable, error))
-        menu->priv->tree = garcon_menu_tree_provider_get_tree (GARCON_MENU_TREE_PROVIDER (merger));
+      if (garcon_menu_merger_run (merger, 
+                                  &menu->priv->merge_files,
+                                  &menu->priv->merge_dirs,
+                                  cancellable, error))
+        {
+          menu->priv->tree = 
+            garcon_menu_tree_provider_get_tree (GARCON_MENU_TREE_PROVIDER (merger));
+        }
       else
-        success = FALSE;
+        {
+          success = FALSE;
+        }
 
       g_object_unref (merger);
     }
@@ -569,14 +733,18 @@ garcon_menu_load (GarconMenu   *menu,
 
   g_object_unref (parser);
 
-  if (G_LIKELY (!success))
+  if (!success)
     return FALSE;
 
   /* Generate submenus */
   garcon_menu_resolve_menus (menu);
 
   /* Resolve the menu directory */
-  garcon_menu_resolve_directory (menu);
+  garcon_menu_resolve_directory (menu, cancellable, TRUE);
+
+  /* Abort if the cancellable was cancelled */
+  if (g_cancellable_set_error_if_cancelled (cancellable, error))
+    return FALSE;
 
   desktop_id_table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
@@ -590,6 +758,9 @@ garcon_menu_load (GarconMenu   *menu,
 
   g_hash_table_unref (desktop_id_table);
 
+  /* Initiate file system monitoring */
+  garcon_menu_start_monitoring (menu);
+
   return TRUE;
 }
 
@@ -745,13 +916,23 @@ garcon_menu_get_directories (GarconMenu *menu)
 
 
 static void
-garcon_menu_resolve_directory (GarconMenu *menu)
+garcon_menu_resolve_directory (GarconMenu   *menu,
+                               GCancellable *cancellable,
+                               gboolean      recursive)
 {
   GarconMenuDirectory *directory = NULL;
   GList               *directories = NULL;
   GList               *iter;
 
   g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
+
+  /* release the old directory if there is one */
+  if (menu->priv->directory != NULL) 
+    {
+      g_object_unref (menu->priv->directory);
+      menu->priv->directory = NULL;
+    }
 
   /* Determine all directories for this menu */
   directories = garcon_menu_get_directories (menu);
@@ -772,9 +953,12 @@ garcon_menu_resolve_directory (GarconMenu *menu)
   /* Free reverse list copy */
   g_list_free (directories);
 
-  /* Resolve directories of submenus recursively */
-  for (iter = menu->priv->submenus; iter != NULL; iter = g_list_next (iter))
-    garcon_menu_resolve_directory (iter->data);
+  if (recursive)
+    {
+      /* Resolve directories of submenus recursively */
+      for (iter = menu->priv->submenus; iter != NULL; iter = g_list_next (iter))
+        garcon_menu_resolve_directory (iter->data, cancellable, recursive);
+    }
 }
 
 
@@ -843,31 +1027,30 @@ garcon_menu_lookup_directory (GarconMenu  *menu,
 
 
 static GList *
-garcon_menu_get_app_dirs (GarconMenu *menu)
+garcon_menu_get_app_dirs (GarconMenu *menu,
+                          gboolean    recursive)
 {
   GList *dirs = NULL;
+  GList *lp;
+  GList *sp;
+  GList *submenu_app_dirs;
 
   /* Fetch all application directories */
   dirs = garcon_menu_node_tree_get_string_children (menu->priv->tree,
                                                     GARCON_MENU_NODE_TYPE_APP_DIR,
                                                     TRUE);
 
-#if 0
-  /* A submenu always inherits the application directories of its parent,
-   * that is the reason the call below was added.
-   * It only turned out we were looking in that same directories for
-   * .desktop files multiple times.
-   *
-   * This was caused by the combination of the parent call below and
-   * traversing the children in garcon_menu_collect_files(). For each
-   * submenu the appdirs of the root were added and traversed again.
-   *
-   * This is not needed because we always start at the root and traverse
-   * in "pre-order", so all the desktop files are added in the hash-table.
-   */
-  if (menu->priv->parent != NULL)
-    dirs = g_list_concat (dirs, garcon_menu_get_app_dirs (menu->priv->parent));
-#endif
+  if (recursive)
+    {
+      for (lp = menu->priv->submenus; lp != NULL; lp = lp->next)
+        {
+          submenu_app_dirs = garcon_menu_get_app_dirs (lp->data, recursive);
+
+          for (sp = g_list_last (submenu_app_dirs); sp != NULL; sp = sp->prev) 
+            if (g_list_find_custom (dirs, sp->data, (GCompareFunc) g_strcmp0) == NULL)
+              dirs = g_list_prepend (dirs, sp->data);
+        }
+    }
 
   return dirs;
 }
@@ -884,7 +1067,7 @@ garcon_menu_collect_files (GarconMenu *menu,
 
   g_return_if_fail (GARCON_IS_MENU (menu));
 
-  app_dirs = garcon_menu_get_app_dirs (menu);
+  app_dirs = garcon_menu_get_app_dirs (menu, FALSE);
 
   /* Collect desktop entry filenames */
   for (iter = app_dirs; iter != NULL; iter = g_list_next (iter))
@@ -1552,3 +1735,549 @@ garcon_menu_get_element_no_display (GarconMenuElement *element)
   else
     return garcon_menu_directory_get_no_display (menu->priv->directory);
 }
+
+
+
+static gboolean
+garcon_menu_get_element_equal (GarconMenuElement *element,
+                               GarconMenuElement *other)
+{
+  g_return_val_if_fail (GARCON_IS_MENU (element), FALSE);
+  g_return_val_if_fail (GARCON_IS_MENU (other), FALSE);
+
+  return GARCON_MENU (element) == GARCON_MENU (other);
+}
+
+
+
+static void
+garcon_menu_start_monitoring (GarconMenu *menu)
+{
+  GList *lp;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  /* Let only the root menu monitor menu files, merge files/directories and app dirs */
+  if (menu->priv->parent == NULL)
+    {
+      garcon_menu_monitor_menu_files (menu);
+
+      garcon_menu_monitor_files (menu, menu->priv->merge_files, 
+                                 garcon_menu_merge_file_changed);
+      
+      garcon_menu_monitor_files (menu, menu->priv->merge_dirs,
+                                 garcon_menu_merge_dir_changed);
+
+      garcon_menu_monitor_app_dirs (menu);
+    }
+
+  garcon_menu_monitor_directory_dirs (menu);
+
+  /* Recurse into submenus */
+  for (lp = menu->priv->submenus; lp != NULL; lp = lp->next)
+    garcon_menu_start_monitoring (lp->data);
+
+}
+
+
+
+static void
+garcon_menu_stop_monitoring (GarconMenu *menu)
+{
+  GList *lp;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  /* Recurse into submenus */
+  for (lp = menu->priv->submenus; lp != NULL; lp = lp->next)
+    garcon_menu_stop_monitoring (lp->data);
+
+  /* Disconnect and destroy all monitors */
+  for (lp = menu->priv->monitors; lp != NULL; lp = lp->next)
+    {
+      g_signal_handlers_disconnect_matched (lp->data, G_SIGNAL_MATCH_DATA,
+                                            0, 0, NULL, NULL, menu);
+      g_object_unref (lp->data);
+    }
+
+  /* Free the monitor list */
+  g_list_free (menu->priv->monitors);
+  menu->priv->monitors = NULL;
+}
+
+
+
+static void
+garcon_menu_monitor_menu_files (GarconMenu *menu)
+{
+  GFileMonitor *monitor;
+  GFile        *file;
+  gchar       **paths;
+  guint         n;
+  gint          i;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  if (menu->priv->uses_custom_path)
+    {
+      /* Monitor the root .menu file */
+      monitor = g_file_monitor (menu->priv->file, G_FILE_MONITOR_NONE, NULL, NULL);
+      if (monitor != NULL)
+        {
+          menu->priv->monitors = g_list_prepend (menu->priv->monitors, monitor);
+          g_signal_connect_swapped (monitor, "changed", 
+                                    G_CALLBACK (garcon_menu_file_changed), menu);
+        }
+    }
+  else
+    {
+      /* Monitor all application menu candidates */
+      for (n = 0; n < G_N_ELEMENTS (GARCON_MENU_ROOT_SPECS); ++n)
+        {
+          paths = garcon_config_build_paths (GARCON_MENU_ROOT_SPECS[n]);
+
+          for (i = g_strv_length (paths)-1; paths != NULL && i >= 0; --i)
+            {
+              file = g_file_new_for_path (paths[i]);
+              
+              monitor = g_file_monitor (file, G_FILE_MONITOR_NONE, NULL, NULL);
+              if (monitor != NULL)
+                {
+                  menu->priv->monitors = g_list_prepend (menu->priv->monitors, monitor);
+                  g_signal_connect_swapped (monitor, "changed",
+                                            G_CALLBACK (garcon_menu_file_changed), menu);
+                }
+
+              g_object_unref (file);
+            }
+
+          g_strfreev (paths);
+        }
+    }
+}
+
+
+
+static gint
+find_file_monitor (GFileMonitor *monitor,
+                   GFile        *file)
+{
+  GFile *monitored_file;
+
+  monitored_file = g_object_get_qdata (G_OBJECT (monitor), garcon_menu_file_quark);
+
+  if (monitored_file != NULL && g_file_equal (monitored_file, file))
+    return 0;
+  else
+    return -1;
+}
+
+
+
+static void
+garcon_menu_monitor_files (GarconMenu *menu,
+                           GList      *files,
+                           gpointer    callback)
+{
+  GFileMonitor *monitor;
+  GList        *lp;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (menu->priv->parent == NULL);
+
+  /* Monitor all files from the list */
+  for (lp = files; lp != NULL; lp = lp->next)
+    {
+      /* Monitor files only if they are not being monitored already */
+      if (g_list_find_custom (menu->priv->monitors, lp->data, 
+                              (GCompareFunc) find_file_monitor) == NULL)
+        {
+          /* Try to monitor the file */
+          monitor = g_file_monitor (lp->data, G_FILE_MONITOR_NONE, NULL, NULL);
+          if (monitor != NULL)
+            {
+              /* Associate the monitor with the monitored file */
+              g_object_set_qdata_full (G_OBJECT (monitor), garcon_menu_file_quark,
+                                       g_object_ref (lp->data), g_object_unref);
+
+              /* Add the monitor to the list of monitors belonging to the menu */
+              menu->priv->monitors = g_list_prepend (menu->priv->monitors, monitor);
+
+              /* Make sure we are notified when the file changes */
+              g_signal_connect_swapped (monitor, "changed", G_CALLBACK (callback), menu);
+            }
+        }
+    }
+}
+
+
+
+static void
+garcon_menu_monitor_app_dirs (GarconMenu *menu)
+{
+  GFile *dir;
+  GList *app_dirs;
+  GList *dirs = NULL;
+  GList *lp;
+  
+  g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (menu->priv->parent == NULL);
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  /* Determine all application directories we are interested in for this menu */
+  app_dirs = garcon_menu_get_app_dirs (menu, TRUE);
+  
+  /* Transform app dir filenames into GFile objects, resolving filenames 
+   * relative to the menu file itself */
+  for (lp = app_dirs; lp != NULL; lp = lp->next)
+    {
+      dir = _garcon_file_new_relative_to_file (lp->data, menu->priv->file);
+      dirs = g_list_prepend (dirs, dir);
+    }
+
+  /* Monitor the app dirs */
+  garcon_menu_monitor_files (menu, dirs, garcon_menu_app_dir_changed);
+
+  /* Release the allocated GFiles and free the list */
+  g_list_foreach (dirs, (GFunc) g_object_unref, NULL);
+  g_list_free (dirs);
+
+  /* Free app dir list */
+  g_list_free (app_dirs);
+}
+
+
+
+static void
+garcon_menu_monitor_directory_dirs (GarconMenu *menu)
+{
+  GFileMonitor *monitor;
+  GFile        *file;
+  GFile        *dir;
+  GList        *directory_files;
+  GList        *directory_dirs;
+  GList        *dp;
+  GList        *lp;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  /* Determine all .directory files we are interested in for this menu */
+  directory_files = garcon_menu_get_directories (menu);
+
+  /* Determine all .directory lookup dirs for this menu */
+  directory_dirs = garcon_menu_get_directory_dirs (menu);
+
+  /* Monitor potential .directory files */
+  for (lp = directory_files; lp != NULL; lp = lp->next)
+    {
+      for (dp = directory_dirs; dp != NULL; dp = dp->next)
+        {
+          dir = _garcon_file_new_relative_to_file (dp->data, menu->priv->file);
+          file = _garcon_file_new_relative_to_file (lp->data, dir);
+
+          /* Only try to monitor the .directory file if we don't do that already */
+          if (g_list_find_custom (menu->priv->monitors, file, 
+                                  (GCompareFunc) find_file_monitor) == NULL)
+            {
+              /* Try to monitor the file */
+              monitor = g_file_monitor (file, G_FILE_MONITOR_NONE, NULL, NULL);
+              if (monitor != NULL)
+                {
+                  /* Associate the monitor with the monitored file */
+                  g_object_set_qdata_full (G_OBJECT (monitor), garcon_menu_file_quark,
+                                           g_object_ref (file), g_object_unref);
+
+                  /* Add the monitor to the list of monitors belonging to the menu */
+                  menu->priv->monitors = g_list_prepend (menu->priv->monitors, monitor);
+
+                  /* Make sure we are notified when the file changes */
+                  g_signal_connect_swapped (monitor, "changed",
+                                            G_CALLBACK (garcon_menu_directory_file_changed),
+                                            menu);
+                }
+            }
+
+          g_object_unref (file);
+          g_object_unref (dir);
+        }
+    }
+
+  /* Free lists */
+  g_list_free (directory_dirs);
+  g_list_free (directory_files);
+}
+
+
+
+static void
+garcon_menu_file_changed (GarconMenu       *menu,
+                          GFile            *file,
+                          GFile            *other_file,
+                          GFileMonitorEvent event_type,
+                          GFileMonitor     *monitor)
+{
+  gboolean higher_priority = FALSE;
+  gboolean lower_priority = FALSE;
+  GFile   *menu_file;
+  gchar  **paths;
+  guint    n;
+  guint    i;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (menu->priv->parent == NULL);
+
+  /* Quick check: reloading is needed if the menu file being used has changed */
+  if (g_file_equal (menu->priv->file, file))
+    {
+      g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+      return;
+    }
+
+  /* Check if the event file has higher priority than the file currently being used */
+  for (n = 0; !lower_priority && !higher_priority && n < G_N_ELEMENTS (GARCON_MENU_ROOT_SPECS); ++n)
+    {
+      /* Get XDG config paths for the root spec (e.g. menus/xfce-applications.menu) */
+      paths = garcon_config_build_paths (GARCON_MENU_ROOT_SPECS[n]);
+
+      for (i = 0; !higher_priority && paths != NULL && paths[i] != NULL; ++i) 
+        {
+          menu_file = g_file_new_for_path (paths[i]);
+
+          if (g_file_equal (menu_file, menu->priv->file))
+            {
+              /* the menu's file comes before the changed file in the load
+               * priority order, so the changed file has a lower priority */
+              lower_priority = TRUE;
+            }
+          else if (g_file_equal (menu_file, file))
+            {
+              /* the changed file comes before the menu's file in the load
+               * priority order, so the changed file has a higher priority */
+              higher_priority = TRUE;
+            }
+
+          g_object_unref (menu_file);
+        }
+
+      g_strfreev (paths);
+    }
+
+  /* If the event file has higher priority, a menu reload is needed */
+  if (!lower_priority && higher_priority)
+    g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+}
+
+
+
+static void
+garcon_menu_merge_file_changed (GarconMenu       *menu,
+                                GFile            *file,
+                                GFile            *other_file,
+                                GFileMonitorEvent event_type,
+                                GFileMonitor     *monitor)
+{
+  g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (menu->priv->parent == NULL);
+
+  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+}
+
+
+
+static void
+garcon_menu_merge_dir_changed (GarconMenu       *menu,
+                               GFile            *file,
+                               GFile            *other_file,
+                               GFileMonitorEvent event_type,
+                               GFileMonitor     *monitor)
+{
+  g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (menu->priv->parent == NULL);
+
+  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+}
+
+
+
+static void
+garcon_menu_app_dir_changed (GarconMenu       *menu,
+                             GFile            *file,
+                             GFile            *other_file,
+                             GFileMonitorEvent event_type,
+                             GFileMonitor     *monitor)
+{
+  GarconMenuItem *item;
+  GFileType       file_type;
+  gboolean        affects_the_outside = FALSE;
+  gchar          *path;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+  g_return_if_fail (menu->priv->parent == NULL);
+
+  if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT)
+    {
+      /* query the type of the changed file */
+      file_type = g_file_query_file_type (file, G_FILE_QUERY_INFO_NONE, NULL);
+
+      if (file_type == G_FILE_TYPE_DIRECTORY)
+        {
+          /* in this situation, an app dir could have 
+           * - become unreadable for the current user
+           * - been deleted
+           * - created (possibly inside an existing one)
+           * this is not trivial to handle, so we simply enforce a
+           * menu reload to deal with the changes */
+          g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+        }
+      else
+        {
+          path = g_file_get_path (file);
+          if (path != NULL && g_str_has_suffix (path, ".desktop"))
+            {
+              /* a regular file changed, try to find the corresponding menu item */
+              item = garcon_menu_find_file_item (menu, file);
+              if (item != NULL)
+                {
+                  /* try to reload the item */
+                  if (garcon_menu_item_reload (item, &affects_the_outside, NULL))
+                    {
+                      if (affects_the_outside)
+                        {
+                          /* if the categories changed, the item might have to be
+                           * moved around between different menus. this is slightly
+                           * more complicated than one would first think, so just
+                           * enforce a complete menu reload for now */
+                          g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+                        }
+                      else
+                        {
+                          /* remove the item from the desktop item cache so we are forced
+                           * to reload it from disk the next time */
+                          garcon_menu_item_cache_invalidate_file (menu->priv->cache, file);
+
+                          /* nothing else to do here. the item should emit a 'changed'
+                           * signal to which users of this library can react */
+                        }
+                    }
+                  else
+                    {
+                      /* remove the item from the desktop item cache so we are forced
+                       * to reload it from disk the next time */
+                      garcon_menu_item_cache_invalidate_file (menu->priv->cache, file);
+
+                      /* failed to reload the menu item. this can have many reasons,
+                       * one of them being that the file permissions might have changed
+                       * or that the file was deleted. handling most situations can be very
+                       * tricky, so, again, we just enfore a menu reload until we have 
+                       * something better */
+                      g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+                    }
+                }
+              else
+                {
+                  /* there could be a lot of stuff happening here. seriously, this
+                   * stuff is complicated. for now, simply enforce a complete reload 
+                   * of the menu structure */
+                  g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+                }
+            }
+          g_free (path);
+        }
+    }
+  else if (event_type == G_FILE_MONITOR_EVENT_DELETED)
+    {
+      /* query the type of the changed file */
+      file_type = g_file_query_file_type (file, G_FILE_QUERY_INFO_NONE, NULL);
+
+      if (file_type == G_FILE_TYPE_DIRECTORY)
+        {
+          /* an existing app dir (or a subdirectory) has been deleted. we
+           * could remove all the items that are in use and reside inside
+           * this root directory. but for now... enforce a menu reload! */
+          g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+        }
+      else 
+        {
+          /* a regular file was deleted, try to find the corresponding menu item */
+          item = garcon_menu_find_file_item (menu, file);
+          if (item != NULL)
+            {
+              /* remove the item from the desktop item cache so we are forced
+               * to reload it from disk the next time */
+              garcon_menu_item_cache_invalidate_file (menu->priv->cache, file);
+
+              /* ok, so a .desktop file was removed. of course we don't know
+               * yet whether there is a replacement in another app dir
+               * with lower priority. we could try to find out but for now
+               * it's easier to simply enforce a menu reload */
+              g_signal_emit (menu, menu_signals[RELOAD_REQUIRED], 0);
+            }
+          else
+            {
+              /* the deleted file hasn't been in use anyway, so removing it
+               * doesn't change anything. so we have nothing to do for a
+               * change, no f****ing menu reload! */
+            }
+        }
+    }
+}
+
+
+
+static void
+garcon_menu_directory_file_changed (GarconMenu       *menu,
+                                    GFile            *file,
+                                    GFile            *other_file,
+                                    GFileMonitorEvent event_type,
+                                    GFileMonitor     *monitor)
+{
+  GarconMenuDirectory *old_directory = NULL;
+
+  g_return_if_fail (GARCON_IS_MENU (menu));
+
+  if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT
+      || event_type == G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED
+      || event_type == G_FILE_MONITOR_EVENT_DELETED 
+      || event_type == G_FILE_MONITOR_EVENT_CREATED)
+    {
+      /* take a reference on the current menu directory */
+      if (menu->priv->directory != NULL)
+        old_directory = g_object_ref (menu->priv->directory);
+                  
+      /* reset the menu directory of the menu and load a new one */
+      garcon_menu_resolve_directory (menu, NULL, FALSE);
+
+      /* notify listeners about the old and new menu directories */
+      g_signal_emit (menu, menu_signals[DIRECTORY_CHANGED], 0, 
+                     old_directory, menu->priv->directory);
+
+      /* release the old menu directory we no longer need */
+      if (old_directory != NULL)
+        g_object_unref (old_directory);
+    }
+}
+
+
+
+static GarconMenuItem *
+garcon_menu_find_file_item (GarconMenu *menu,
+                            GFile      *file)
+{
+  GarconMenuItem *item = NULL;
+  GList          *lp;
+
+  g_return_val_if_fail (GARCON_IS_MENU (menu), NULL);
+  g_return_val_if_fail (G_IS_FILE (file), NULL);
+
+  item = garcon_menu_item_pool_lookup_file (menu->priv->pool, file);
+
+  if (item == NULL)
+    {
+      for (lp = menu->priv->submenus; item == NULL && lp != NULL; lp = lp->next)
+        item = garcon_menu_find_file_item (lp->data, file);
+    }
+
+  return item;
+
+}
